/*
 * Copyright (c) 1998 by Aaron Wallack, Natick, MA USA
 * All rights reserved. This material contains unpublished,
 * copyrighted work, which includes confidential and proprietary
 * information of Aaron Wallack.
 */

#ifndef linutil_h
#define linutil_h

#include <math.h>
#include "complex.h"
#include "matrix.h"

#ifndef FALSE
#define FALSE 0
#endif

#ifndef NULL
#define NULL 0
#endif

#ifndef TRUE
#define TRUE 1
#endif

/* Note:

   Arrays:

   Doubly indexed arrays are used to characterize matrices. We chose
   to use doubly indexed arrays because they provide flexibility for
   variable sized arrays.

   Matrices in this C library are built using pointers to doubly
   indexed arrays. In matrix.h, the mat element of the matrix struct
   is a double-pointer to a double "double **mat". 

   +------------------+
   | matrix Operation |
   +------------------+
   	    |
   	    |
	    V
   +------------------+
   |    pointer op    |
   +------------------+

   If you look at the code, most of the Matrix member function calls
   (zeroMatrix, freeMatrix) call the associated Array member functions
   (zeroArray, freeArray).

   One of the drawbacks of doubly indexed arrays is
   that the [i][j] element corresponds to [row][col], which is (y,x),
   and different than the intuitive nomenclature of (x,y), or
   [col][row]. 
 
   Consequently, even though doubly indexed arrays are usually
   specified by (width, height), the inner loops are written so that
   they process rows at a time, i.e.

   for (y = 0; y < ht; ++y)
     for (x = 0; x < wd; ++x)
       ...

  Windows:

  Windows of arrays correspond to a subblock of an array:

  Suppose A refers to the following matrix:
  [  a   b   c   d  ]
  [  e   f   g   h  ]
  [  i   j   k   l  ]
  [  m   n   o   p  ]

  The window B could refer to the subblock including the second and
  third columns and rows:

  [  f   g  ]
  [  j   k  ]

  Windows share the matrix elements with the original doubly indexed
  array; in other words, if we change B[0][0] from f to z, then
  A[1][1]'s element would also change from f to z. The main reason we
  define window objects is so that we can easily copy subblocks of the
  matrices.

  Inevitably, we need to treat allocating and freeing windows
  differently than arrays because we don't want to deallocate the
  storage space (since it's shared). Consequently, we need to call
  windowArray to specify a window and freeWindowArray to deallocate
  it. 

  Note that "windows" can be used just like "arrays" for all matrix
  processing operations (adding arrays, multiplying arrays, ...). The
  only differences are in their construction and destruction */

#ifdef GCC_PROTOTYPES
double *createArray(double **M, int w, int h);
/* returns 	a linear array of doubles corresponding to
   		concatenating the rows of M
   		[ < row 1 > < row 2 > < row 3 > ... ]
*/

void printArray(double **M, int w, int h);
/* effect	prints out the given doubly indexed array
 */

void convertArray(double *L,double **M, int w, int h);
/* effect	unwraps the linear array "L" and stores it into M
   		i.e. assumes L corresponds to concatenating the rows
		of some w x h matrix Q, and effectively copies Q to M
		so that M[i][j] = L[i*w+j]
*/
		
complex *createComplexArray(double **M, int w, int h);
/* returns 	a linear array of complex objects corresponding to
   		concatenating the rows of M
   		[ < row 1 > < row 2 > < row 3 > ... ]
*/

double **eyeArray(int n);
/* returns	a doubly indexed array corresponding to the n x n
		identity matrix
*/

double **zerosArray(int w, int h);
/* returns	a doubly indexed array corresponding to the w x h
		zero matrix
*/

double **makeLinearArray(int n);
/* returns	a doubly indexed array corresponding to the n x n
		integer matrix:
		[ 1 0 0 0 ... ]
		[ 0 2 0 0 ... ]
		[ 0 0 3 0 ... ]
		...
*/

void freeArray(double **M, int h);
/* effect	deallocates the doubly indexed array M
 */

void freeWindowArray(double **M);
/* effect	deallocates the doubly indexed array window M
 */

void addArray(double **M1, double **M2, double **DST, int w, int h);
/* effect	Sets the elements in DST to be the sum of the
   		corresponding elements in M1 and M2, i.e., 
		DST[i][j] = M1[i][j] + M2[i][j]
*/

void subtractArray(double **M1, double **M2, double **DST, int w, int h);
/* effect	Sets the elements in DST to be the difference of the
   		corresponding elements in M1 and M2, i.e., 
		DST[i][j] = M1[i][j] - M2[i][j]
*/

void multiplyArray(double **M1, double **M2, double **DST, int w, int
		   h, int w2);
/* effect	Sets the elements in DST to be the matrix product of the
   		corresponding rows in M1 and columns in M2, i.e., 
		DST[i][j] = Sum_{k=1..w} M1[i][k] * M2[k][j]
*/

void copyArray(double **SRC, double **DST, int w, int h);
/* effect	Sets the elements in DST to be equal to 
   		corresponding elemnts in SRC, i.e., 
		DST[i][j] = SRC[i][j]
*/

void subArray(double **SRC, double **DST, int x, int y, int w, int h);
/* effect	Sets the elements in DST to be equal to 
   		windowed elemnts in SRC, i.e., 
		DST[i][j] = SRC[i+x][j+y]
*/

double mypower(double val, double exponent);
/* returns	val^exponent      if val >= 0
   		-(-val^exponent)  if val < 0
*/

void windowArray(double **SRC, double **DST, int x, int y, int w, int h);
/* effect	Make DST a window array of SRC
 */

void negativeArray(double **SRC, double **DST, int w, int h);
/* effect	Sets the elements in DST to be opposite of
   		corresponding elemnts in SRC, i.e., 
		DST[i][j] = -SRC[i][j]
*/

void scaleArray(double **M1, double s, double **DST, int w, int h);
/* effect	Sets the elements in DST to be multiples of
   		corresponding elemnts in M1, i.e., 
		DST[i][j] = s * M1[i][j]
*/

#else

double *createArray();
complex *createComplexArray();
complex *createComplexMatrix();
double **zerosArray();
double **eyeArray();
double **makeLinearArray();
void freeArray();
void copyArray();
void subArray();
void windowArray();
void negativeArray();
double mypower();

#endif

#endif /* linutil_h */
