/*
 * Copyright (c) 1998 by Aaron Wallack, Natick, MA USA
 * All rights reserved. This material contains unpublished,
 * copyrighted work, which includes confidential and proprietary
 * information of Aaron Wallack.
 */

#ifndef matrix_h
#define matrix_h

#include "linutil.h"

/* matrix class characterizes matrices of double precision values
   The "mat" member is a doubly indexed array of doubles
   and mat[row][column] corresponds to the element in the specified
   row and column. The matrix object also knows about the width
   and height of the matrix

  Windows:

  Windows of arrays correspond to a subblock of an array:

  Suppose A refers to the following matrix:
  [  a   b   c   d  ]
  [  e   f   g   h  ]
  [  i   j   k   l  ]
  [  m   n   o   p  ]

  The window B could refer to the subblock including the second and
  third columns and rows:

  [  f   g  ]
  [  j   k  ]

  Windows share the matrix elements with the original matrix; in other
  words, if we change B[0][0] from f to z, then A[1][1]'s element
  would also change from f to z. The main reason we define
  window objects is so that we can easily copy subblocks of the
  matrices. 

  Inevitably, we need to treat allocating and freeing windows
  differently than arrays because we don't want to deallocate the
  storage space (since it's shared). Consequently, we need to call
  windowMatrix to specify a window and freeWindowMatrix to deallocate
  it. 

  Note that "windows" can be used just like "matrice" for all matrix
  processing operations (adding matrices, multiplying matrices, ...). The
  only differences are in their construction and destruction

*/

typedef struct matrix
{
  double **mat; /* Doubly indexed pointer to 2D array of doubles*/
  int width;    /* width of matrix */
  int height;   /* height of matrix */
} matrix;

/* Note that member functions all use the suffix Matrix to
   distinguish that they act upon matrix objects
*/

#ifdef GCC_PROTOTYPES
void printMatrix(matrix *mat);
/* effect	print the matrix to stdout
 */

matrix *zerosMatrix(int w, int h);
/* returns	a newly allocated matrix object characterizing a w x h 
   		matrix
*/

matrix *eyeMatrix(int n);
/* returns	a newly allocated matrix object characterizing an n x n
   		identity matrix
*/

matrix *makeLinearMatrix(int n);
/* returns	a newly allocated matrix object characterizing an n x n
   		integer matrix

		integer matrix:
		[ 1 0 0 0 ... ]
		[ 0 2 0 0 ... ]
		[ 0 0 3 0 ... ]
		...
*/

void freeMatrix(matrix *mat);
/* effect	deallocate matrix object mat
 */

matrix *addMatrix(matrix *m1, matrix *m2, matrix *dst);
/* returns	dst
   effect	make dst the sum of m1 and m2
   requires	m1 and m2 are the same size
   		dst and m1 are the same size (if dst != 0)
   note		a new dst matrix is allocated and returned if dst = 0
*/

matrix *copyMatrix(matrix *src, matrix *dst);
/* returns	dst
   effect	make dst the same as src
   requires	dst and src are the same size (if dst != 0)
   note		a new dst matrix is allocated and returned if dst = 0
*/

matrix *subMatrix(matrix *src, matrix *dst, int x, int y, int w, int h);
/* returns	dst
   effect	make dst a copy of window src from x,y to x+w,y+h
   note		a new dst matrix is allocated and returned if dst = 0
*/

matrix *subtractMatrix(matrix *m1, matrix *m2, matrix *dst);
/* returns	dst
   effect	make dst the difference between m1 and m2
   requires	m1 and m2 are the same size
   		dst and m1 are the same size (if dst != 0)
   note		a new dst matrix is allocated and returned if dst = 0
*/

matrix *multiplyMatrix(matrix *m1, matrix *m2, matrix *dst);
/* returns	dst
   effect	make dst the matrix product of m1 and m2
   requires	m1.width == m2.height
   		dst.width == m2.width and dst.height == m1.height (if dst != 0)
   note		a new dst matrix is allocated and returned if dst = 0
*/

matrix *resizeMatrix(matrix *m, int w, int h);
/* returns	m
   effect	resizes matrix m to have width w and height h
   requires	m is bound, i.e. m != 0
*/

matrix *inverseMatrix(matrix *m, matrix *dst);
/* returns	dst
   effect	make dst the multiplicative inverse of m, i.e., dst * m == I
   requires	dst and m are the same size (if dst != 0)
   note		a new dst matrix is allocated and returned if dst = 0
*/

matrix *windowMatrix(matrix *src, matrix *dst, int x, int y, int w, int h);
/* returns	dst
   effect	make dst the matrix window corresponding to
   		x,y..x+w,y+h in src
   requires	x > 0, y > 0, w > 0, h > 0
   		x+w < src.width, y+h < src.height
   note		a new dst window matrix is allocated and returned if dst = 0

*/

void freeWindowMatrix(matrix *mat);
/* effect	Free a window matrix object (must be created by
   		calling windowMatrix with dst==0
*/

matrix *negativeMatrix(matrix *m, matrix *dst);
/* returns	dst
   effect	make dst the additive inverse of m, i.e., dst + m == 0
   requires	dst and m are the same size (if dst != 0)
   note		a new dst matrix is allocated and returned if dst = 0
*/

matrix *scaleMatrix(matrix *m, double s, matrix *dst);
/* returns	dst
   effect	make dst the scale of m, i.e., dst = m * s
   requires	dst and m are the same size (if dst != 0)
   note		a new dst matrix is allocated and returned if dst = 0
*/
   
#else
matrix *zerosMatrix();
matrix *eyeMatrix();
matrix *makeLinearMatrix();
matrix *subMatrix();
matrix *inverseMatrix();
matrix *multiplyMatrix();
matrix *windowMatrix();
void freeWindowMatrix();
matrix *negativeMatrix();
matrix *scaleMatrix();
void freeMatrix();
#endif

#endif /* matrix_h */
